from logging import Logger, getLogger
from typing import List

from docutils.nodes import Node, SkipNode, Text, literal_block, system_message
from docutils.utils import Reporter
from sphinx.writers.text import MAXWIDTH
from sphinxcontrib.writers.rst import RstTranslator, RstWriter

from bpystubgen.nodes import APIMember, Argument, AttributeRef, Class, ClassRef, Data, DataRef, DocString, Function, \
    FunctionRef, FunctionScope, Import, MethodRef, Module, ModuleRef, Property, PropertyRef, Reference


class StubWriter(RstWriter):

    def translate(self) -> None:
        visitor = StubTranslator(self.document, self.builder)

        self.document.walkabout(visitor)
        self.output = visitor.body


# noinspection PyPep8Naming, PyUnusedLocal
class StubTranslator(RstTranslator):
    logger: Logger = getLogger("StubTranslator")

    def wrap(self, text: str, width: int = MAXWIDTH) -> List[str]:
        # Disable wrapping.
        return [text]

    def visit_DocString(self, node: DocString) -> None:
        self.new_state(0)
        self.add_text("\"\"\"")

    def depart_DocString(self, node: DocString) -> None:
        self.add_text("\"\"\"")
        self.end_state()

    def visit_Module(self, node: Module) -> None:
        self.new_state(0)

    def depart_Module(self, node: Module) -> None:
        self.end_state()

    def visit_Import(self, node: Import) -> None:
        self.new_state(0)
        self.add_text(node.astext())
        self.end_state()

        raise SkipNode

    def depart_Import(self, node: Import) -> None:
        pass

    def visit_APIMember(self, node: APIMember) -> None:
        name = node.name

        if not name or not any(name):
            self.logger.warning("Ignoring a member node without a name: %s", node.astext())
            raise SkipNode

        self.new_state(0)

        for line in node.signature.split("\n"):
            self.new_state(0)
            self.add_text(line)
            self.end_state()

        if node.has_body:
            self.new_state(self.indent)

    def depart_APIMember(self, node: APIMember) -> None:
        if node.has_body:
            self.end_state()

        self.end_state()

    def visit_Data(self, node: Data) -> None:
        self.visit_APIMember(node)

    def depart_Data(self, node: Data) -> None:
        self.depart_APIMember(node)

    def visit_Property(self, node: Property) -> None:
        self.visit_APIMember(node)

    def depart_Property(self, node: Property) -> None:
        self.add_text("...")
        self.depart_APIMember(node)

    def visit_Function(self, node: Function) -> None:
        # Fix documentation errors.
        if isinstance(node.parent, Class) and node.scope == FunctionScope.Module:
            node.scope = FunctionScope.Instance

        self.visit_APIMember(node)

    def depart_Function(self, node: Function) -> None:
        self.add_text("...")
        self.depart_APIMember(node)

    def visit_Class(self, node: Class) -> None:
        self.visit_APIMember(node)

    def depart_Class(self, node: Class) -> None:
        if not any(node.members):
            self.add_text("...")

        self.depart_APIMember(node)

    def visit_Argument(self, node: Argument) -> None:
        raise SkipNode

    def depart_Argument(self, node: Argument) -> None:
        pass

    def visit_Reference(self, node: Reference) -> None:
        self.add_text(node.astext())

        raise SkipNode

    def visit_ModuleRef(self, node: ModuleRef) -> None:
        self.visit_Reference(node)

    def depart_ModuleRef(self, node: ModuleRef) -> None:
        pass

    def visit_DataRef(self, node: DataRef) -> None:
        self.visit_Reference(node)

    def depart_DataRef(self, node: DataRef) -> None:
        pass

    def visit_FunctionRef(self, node: FunctionRef) -> None:
        self.visit_Reference(node)

    def depart_FunctionRef(self, node: FunctionRef) -> None:
        pass

    def visit_ClassRef(self, node: ClassRef) -> None:
        self.visit_Reference(node)

    def depart_ClassRef(self, node: ClassRef) -> None:
        pass

    def visit_AttributeRef(self, node: AttributeRef) -> None:
        self.visit_Reference(node)

    def depart_AttributeRef(self, node: AttributeRef) -> None:
        pass

    def visit_PropertyRef(self, node: PropertyRef) -> None:
        self.visit_Reference(node)

    def depart_PropertyRef(self, node: PropertyRef) -> None:
        pass

    def visit_MethodRef(self, node: MethodRef) -> None:
        self.visit_Reference(node)

    def depart_MethodRef(self, node: MethodRef) -> None:
        pass

    def visit_system_message(self, node: system_message) -> None:
        level = node["level"]

        if level == Reporter.DEBUG_LEVEL:
            logger = self.logger.debug
        elif level == Reporter.INFO_LEVEL:
            logger = self.logger.info
        elif level == Reporter.WARNING_LEVEL:
            logger = self.logger.warning
        elif level == Reporter.ERROR_LEVEL:
            logger = self.logger.error
        elif level == Reporter.SEVERE_LEVEL:
            logger = self.logger.fatal
        else:
            logger = self.logger.log

        logger(node.astext())

        raise SkipNode

    def visit_literal_block(self, node: literal_block) -> None:
        for c in tuple(node.children):
            node.remove(c)

        node += Text(node.rawsource.replace('"""', '\\"\\"\\"'))

        super().visit_literal_block(node)

    def unknown_visit(self, node: Node) -> None:
        self.logger.warning("Unknown node type: %s", node)

    def unknown_departure(self, node: Node) -> None:
        pass



import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

from bpystubgen.generator import generate

parser = ArgumentParser(
    prog="bpystubgen",
    description="Generate Python API stubs from Blender's documentation.")

parser.add_argument("input", type=str,
                    help="Source directory where *.rst files are located")
parser.add_argument("output", type=str, default=".",
                    help="Output directory where generated modules will be saved")
parser.add_argument("--verbose", default=False, action="store_true", help="Print debug messages")
parser.add_argument("--quiet", default=False, action="store_true", help="Print only error messages")

args = parser.parse_args()

source = Path(args.input).expanduser()
dest = Path(args.output).expanduser()

if not source.is_dir():
    sys.exit(f"The specified input is not a valid directory: {source}")

if dest.exists():
    if dest.is_file():
        sys.exit(f"The specified output already exists but it's not a valid directory: {dest}")
else:
    dest.mkdir(parents=True)

if args.quiet:
    log_level = logging.WARNING
elif args.verbose:
    log_level = logging.DEBUG
else:
    log_level = logging.INFO

generate(source, dest, log_level)



import re
from itertools import repeat
from typing import Final, Optional

_simple_pattern: Final = re.compile(
    "^(?:unsigned\\s)?(?P<type>[a-zA-Z]+)(?:\\s?\\([^)]+\\))?(?:\\sin\\s[\\[][^]]+[]])?(?:[,.\\s].*)?$")

_container_of_pattern: Final = re.compile(
    "^(?P<container>[a-zA-Z]+)\\sof\\s(?::class:`[~!]?(?P<reference>[^`]+)`|"
    "(?P<data>[a-zA-Z]+))(?P<qualifier>[',.\\s].*)?$")

_prop_collection_of_pattern: Final = re.compile(
    "^:class:`bpy_prop_collection`\\sof\\s(?::class:`[~!]?(?P<reference>[^`]+)`|"
    "(?P<data>[a-zA-Z]+))(?P<qualifier>[',.\\s].*)?$")

_qualified_list_pattern: Final = re.compile(
    "^list\\s*\\([a-zA-Z\\s]*vector\\sof\\s[0-9]+\\s(?P<data>[a-zA-Z]+)(?:[,.\\s][^)]*)?\\)(?:[,.\\s].*)?$")

_list_bracket_pattern: Final = re.compile(
    "^list\\s*\\[(?::class:`[~!]?(?P<reference>[^`]+)`|(?P<data>[a-zA-Z]+))](?:[,.\\s].*)?$")

_array_of_pattern: Final = re.compile(
    "^(?P<data>[a-z]+)\\sarray\\sof\\s(?P<count>[0-9]+)\\sitems.*$")

_multi_array_of_pattern: Final = re.compile(
    "^(?P<data>[a-z]+)\\smulti-dimensional\\sarray\\sof\\s(?P<rows>[0-9]+)\\s*\\*\\s*(?P<cols>[0-9]+)\\sitems.*$")

_dictionary_pattern: Final = re.compile(
    "^dict(?:ionary)?\\s*[\\[(](?P<key>[^,\\s]+)\\s*,\\s*(?P<value>[^])]+)[])].*$")

_reference_type_pattern: Final = re.compile(
    "^:class:`[~!]?(?P<name>[a-zA-Z_0-9.\\s]+)(?:\\s*<(?P<target>[a-zA-Z_0-9.]+)>)?`(?:[,.\\s].*)?$")

_known_types: Final = {
    "any": "typing.Any",
    "str": "str",
    "strs": "str",
    "string": "str",
    "strings": "str",
    "int": "int",
    "ints": "int",
    "integer": "int",
    "integers": "int",
    "float": "float",
    "floats": "float",
    "double": "float",
    "doubles": "float",
    "bool": "bool",
    "bools": "bool",
    "boolean": "bool",
    "booleans": "bool",
    "class": "typing.Type",
    "classes": "typing.Type",
    "type": "typing.Type",
    "types": "typing.Type",
    "object": "bpy.types.Object",
    "objects": "bpy.types.Object",
    "callable": "typing.Callable",
    "callables": "typing.Callable",
    "function": "typing.Callable",
    "functions": "typing.Callable",
    "dict": "typing.Dict[str, typing.Any]",
    "dictionary": "typing.Dict[str, typing.Any]",
    "dictionaries": "typing.Dict[str, typing.Any]",
    "set": "typing.Set[typing.Any]",
    "sets": "typing.Set[typing.Any]",
    "sequence": "typing.Sequence[typing.Any]",
    "sequences": "typing.Sequence[typing.Any]",
    "list": "typing.List[typing.Any]",
    "lists": "typing.List[typing.Any]",
    "array": "typing.Tuple[typing.Any, ...]",
    "arrays": "typing.Tuple[typing.Any, ...]",
    "tuple": "typing.Tuple[typing.Any, ...]",
    "tuples": "typing.Tuple[typing.Any, ...]"
}

_container_types: Final = {
    "list": "typing.List",
    "set": "typing.Set",
    "vector": "typing.List",
    "sequence": "typing.Sequence"
}


def parse_simple(text: str) -> Optional[str]:
    result = _simple_pattern.match(text)

    if result:
        type_info = result.group("type").lower()

        if type_info in _known_types:
            return _known_types[type_info]

    return None


def parse_reference(text: str) -> Optional[str]:
    result = _reference_type_pattern.match(text)

    return result.group("target") or result.group("name") if result else None


def parse_qualified_list(text: str) -> Optional[str]:
    result = _qualified_list_pattern.match(text)

    if not result:
        return None

    data_type = result.group("data")

    if data_type in _known_types:
        return f"typing.List[{_known_types[data_type]}]"

    return None


def parse_bracket_list(text: str) -> Optional[str]:
    result = _list_bracket_pattern.match(text)

    if not result:
        return None

    data_type = result.group("data")

    if data_type in _known_types:
        data_type = _known_types[data_type]

    if not data_type:
        data_type = result.group("reference")

    if data_type:
        return f"typing.List[{data_type}]"

    return None


def parse_prop_collection_of(text: str) -> Optional[str]:
    result = _prop_collection_of_pattern.match(text)

    if not result:
        return None

    data_type = result.group("data")
    qualifier = result.group("qualifier")

    if data_type in _known_types:
        data_type = _known_types[data_type]
    else:
        reference = result.group("reference")
        data_type = parse_reference("".join((":class:`", reference, "`"))) if reference else None

    if not data_type:
        return None

    if qualifier:
        qualifier = qualifier.strip()

        if qualifier.startswith("tuple"):
            data_type = f"typing.Tuple[{data_type}, ...]"
        elif qualifier.startswith("list"):
            data_type = f"typing.List[{data_type}]"
        elif qualifier.startswith("sequence"):
            data_type = f"typing.Sequence[{data_type}]"

    return f"typing.Union[typing.Sequence[{data_type}], typing.Mapping[str, {data_type}]]"


def parse_container_of(text: str) -> Optional[str]:
    result = _container_of_pattern.match(text)

    if not result:
        return None

    container_type = result.group("container")
    data_type = result.group("data")
    qualifier = result.group("qualifier")

    if container_type in _container_types:
        container_type = _container_types[container_type]
    else:
        container_type = None

    if data_type in _known_types:
        data_type = _known_types[data_type]
    else:
        reference = result.group("reference")
        data_type = parse_reference("".join((":class:`", reference, "`"))) if reference else None

    if not data_type or not container_type:
        return None

    if qualifier:
        qualifier = qualifier.strip()

        if qualifier.startswith("tuple"):
            data_type = f"typing.Tuple[{data_type}, ...]"
        elif qualifier.startswith("list"):
            data_type = f"typing.List[{data_type}]"
        elif qualifier.startswith("sequence"):
            data_type = f"typing.Sequence[{data_type}]"

    return f"{container_type}[{data_type}]"


def parse_array_of(text: str) -> Optional[str]:
    result = _array_of_pattern.match(text)

    if not result:
        return None

    data_type = result.group("data")
    count = int(result.group("count"))

    if data_type in _known_types:
        data_type = _known_types[data_type]
    else:
        return None

    if count > 5:
        return f"typing.Tuple[{data_type}, ...]"

    return f"typing.Tuple[{', '.join(repeat(data_type, count))}]"


def parse_multi_array_of(text: str) -> Optional[str]:
    result = _multi_array_of_pattern.match(text)

    if not result:
        return None

    data_type = result.group("data")
    rows = int(result.group("rows"))
    cols = int(result.group("cols"))

    if data_type in _known_types:
        data_type = _known_types[data_type]
    else:
        return None

    def make_tuple(t, n):
        if n > 5:
            return f"typing.Tuple[{t}, ...]"

        return f"typing.Tuple[{', '.join(repeat(t, n))}]"

    return make_tuple(make_tuple(data_type, cols), rows)


def parse_dictionary(text: str) -> Optional[str]:
    result = _dictionary_pattern.match(text)

    if not result:
        return None

    def guess_type(v):
        if not v:
            return "typing.Any"

        if v in _known_types:
            return _known_types[v]

        return parse_reference(v) or "typing.Any"

    key = guess_type(result.group("key"))
    value = guess_type(result.group("value"))

    return f"typing.Dict[{key}, {value}]"


def parse_special_cases(text: str) -> Optional[str]:
    text = text.lower()

    if "enum in" in text:
        return "str"
    elif "enum set in" in text:
        return "typing.Set[str]"

    if text.startswith("function"):
        return "typing.Callable"

    if text.startswith("vector"):
        return "mathutils.Vector"
    else:
        segments = text.split(" ")

        if len(segments) > 1 and segments[1] == "vector":
            return "mathutils.Vector"

    return None


def parse_type(text: str) -> Optional[str]:
    parsers = (
        parse_array_of,
        parse_multi_array_of,
        parse_qualified_list,
        parse_bracket_list,
        parse_container_of,
        parse_prop_collection_of,
        parse_dictionary,
        parse_reference,
        parse_simple,
        parse_special_cases
    )

    try:
        return next(filter(lambda r: r, map(lambda p: p(text), parsers)))
    except StopIteration:
        return None



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from graphlib import TopologicalSorter
from io import StringIO
from typing import Final, Mapping, Optional, Sequence, Set, Tuple, cast

from docutils.nodes import Element, Inline, TextElement

from bpystubgen.parser import _known_types


class Referencing(ABC):

    @property
    def referred_types(self) -> Set[str]:
        return {"typing"}


class Typed(Element, Referencing, ABC):

    @property
    def type(self) -> Optional[str]:
        return self.attributes["type"] if self.hasattr("type") else None

    @type.setter
    def type(self, value: Optional[str]) -> None:
        if value:
            self.attributes["type"] = value
        elif "type" in self.attributes:
            del self.attributes["type"]

    @property
    def referred_types(self) -> Set[str]:
        references = super().referred_types

        rtype = self.type

        if rtype and rtype != "None":
            references.add(rtype)

        return references


class Named(Element, ABC):

    @property
    def name(self) -> Optional[str]:
        return self.attributes["name"] if self.hasattr("name") else None

    @name.setter
    def name(self, value: Optional[str]) -> None:
        if value:
            self.attributes["name"] = value
        elif "name" in self.attributes:
            del self.attributes["name"]


class Documentable(Element, ABC):

    @property
    def docstring(self) -> Optional[DocString]:
        try:
            docstring = next(filter(lambda c: isinstance(c, DocString), self.children))

            return cast(DocString, docstring)
        except StopIteration:
            return None


class Referencable(Named, ABC):

    @abstractmethod
    def create_ref(self) -> Optional[Reference]:
        pass


class APIMember(Documentable, Referencable, Typed, Element, ABC):

    @property
    @abstractmethod
    def signature(self) -> str:
        pass

    @property
    def full_name(self) -> Optional[str]:
        name = self.name

        if not name:
            return None

        identifiers = [name]
        parent = self.parent

        while isinstance(parent, Named):
            name = parent.name

            if name:
                identifiers.append(name)
            else:
                return None

            parent = parent.parent

        identifiers.reverse()

        return ".".join(identifiers)

    @property
    @abstractmethod
    def has_body(self) -> bool:
        pass

    @property
    def module(self) -> Optional[Module]:
        parent = self.parent

        while parent:
            if isinstance(parent, Module):
                return parent

            parent = parent.parent

        return None

    def localise_name(self, name: str) -> str:
        module = self.module

        return module.localise_name(name) if module else name


class APICollection(Element, ABC):

    @property
    def members(self) -> Sequence[APIMember]:
        return tuple(self.traverse(APIMember, include_self=False, ascend=False))


class Module(Referencable, Referencing, Documentable, APICollection):
    tagname = "module"

    def create_ref(self, simple: bool = False) -> Optional[Reference]:
        name = self.name

        if not name:
            return None

        if simple:
            name = "~" + name

        return ModuleRef(text=name)

    @property
    def referred_types(self) -> Set[str]:
        references = super().referred_types

        for member in self.members:
            references = references.union(member.referred_types)

        return references

    @property
    def imports(self) -> Sequence[Import]:
        return tuple(self.traverse(Import, include_self=False, ascend=False))

    def import_types(self):
        types_to_imports = set()

        for tpe in self.referred_types:
            name = tpe.replace('"', "")

            if name.startswith("class:"):
                name = name.split(":")[1] \
                    .replace("`", "") \
                    .replace("!", "") \
                    .replace("~", "")

            if not name[0].islower() or name in _known_types or name.startswith(self.name):
                continue

            types_to_imports.add(name.split(".")[0])

        for i in self.imports:
            i.parent.remove(i)

        index = 1 if self.docstring else 0

        for tpe in types_to_imports:
            self.insert(index, Import(module=tpe))

    def localise_name(self, name: str) -> str:
        prefix = self.name + "."

        if name.startswith(prefix):
            return name[len(prefix):]

        # XXX: Hack to replace types in containers (e.g. typing.List[bge.types.KX_GameObject])
        if prefix in name and "typing" in name:
            for cls in filter(lambda m: isinstance(m, Class), self.members):
                name = name.replace(cls.full_name, cls.name)

        return name

    def sort_members(self) -> None:
        classes = tuple(filter(lambda m: isinstance(m, Class), self.members))

        if not any(classes):
            return

        local_types: Mapping[str, Class] = dict(map(lambda cl: (cl.name, cl), classes))

        def collect_deps(tpe: str, deps: Set[str], add_self: bool = True) -> Set[str]:
            if tpe in local_types:
                if add_self:
                    deps.add(tpe)

                for r in local_types[tpe].referred_types:
                    local_name = self.localise_name(r)
                    deps = deps.union(collect_deps(local_name, deps))

            return deps

        def create_entry(m: APIMember) -> Tuple[str, Set[str]]:
            cls = cast(Class, m)
            types = set(filter(lambda t: t in local_types and t != m.name, map(self.localise_name, cls.base_types)))
            return m.name, types

        graph = dict(map(create_entry, classes))
        ordered = TopologicalSorter(graph).static_order()

        pos = self.children.index(classes[0])

        for ch in classes:
            ch.parent.remove(ch)

        for name in ordered:
            self.insert(pos, local_types[name])
            pos += 1


class Data(APIMember):
    tagname = "data"

    @property
    def has_body(self) -> bool:
        return False

    @property
    def signature(self) -> str:
        name = self.name
        type_info = self.type

        if not name or not any(name):
            raise ValueError("Data node does not have a name.")

        type_info = self.localise_name(type_info) if type_info else "typing.Any"

        return f"{name}: {type_info} = ..."

    def create_ref(self, simple: bool = False) -> Optional[Reference]:
        name = self.full_name

        if not name:
            return None

        if simple:
            name = "~" + name

        return DataRef(text=name)


class Property(APIMember):
    tagname = "data"

    @property
    def has_body(self) -> bool:
        return True

    @property
    def signature(self) -> str:
        name = self.name
        type_info = self.type

        if not name or not any(name):
            raise ValueError("Property node does not have a name.")

        out = StringIO()

        out.write("@property\n")
        out.write("def ")
        out.write(name)
        out.write("(self) -> ")
        out.write(self.localise_name(type_info) if type_info else "typing.Any")
        out.write(":")

        return out.getvalue()

    def create_ref(self, simple: bool = False) -> Optional[Reference]:
        name = self.full_name

        if not name:
            return None

        if simple:
            name = "~" + name

        return PropertyRef(text=name)


class FunctionLike(APIMember, ABC):

    @property
    def arguments(self) -> Sequence[Argument]:
        # noinspection PyTypeChecker
        return tuple(filter(lambda c: isinstance(c, Argument), self.children))

    @property
    def has_body(self) -> bool:
        return True

    @property
    def referred_types(self) -> Set[str]:
        references = super().referred_types

        for arg in self.arguments:
            rtype = arg.type

            if rtype:
                references.add(rtype)

        return references


class Class(FunctionLike, APICollection):
    tagname = "class"

    @property
    def base_types(self) -> Sequence[str]:
        if not self.hasattr("base_types"):
            return ()

        value = str(self.attributes["base_types"]).split(",")

        return tuple(map(lambda v: v.strip(), value))

    @base_types.setter
    def base_types(self, value: Sequence[str]) -> None:
        if value and any(value):
            self.attributes["base_types"] = ", ".join(value)
        elif "base_types" in self.attributes:
            del self.attributes["base_types"]

    @property
    def referred_types(self) -> Set[str]:
        references = super().referred_types

        for member in self.members:
            references = references.union(member.referred_types)

        references = references.union(self.base_types)

        return references

    @property
    def signature(self) -> str:
        name = self.name

        if not name or not any(name):
            raise ValueError("Class node does not have a name.")

        base_types = self.base_types

        if any(base_types):
            parents = ", ".join(map(self.localise_name, base_types))

            return "".join(["class ", name, "(", parents, "):"])
        else:
            return "".join(["class ", name, ":"])

    def create_ref(self, simple: bool = False) -> Optional[Reference]:
        name = self.full_name

        if not name:
            return None

        if simple:
            name = "~" + name

        return ClassRef(text=name)


class FunctionScope(Enum):
    Module: Final = "module"
    Instance: Final = "instance"
    Class: Final = "class"
    Static: Final = "static"


class Function(FunctionLike):
    tagname = "function"

    @property
    def scope(self) -> FunctionScope:
        value = str.capitalize(self.attributes["scope"]) if self.hasattr("scope") else None

        return FunctionScope[value] if value in FunctionScope.__members__ else FunctionScope.Module

    @scope.setter
    def scope(self, value: FunctionScope) -> None:
        if value and value != FunctionScope.Module:
            self.attributes["scope"] = str.lower(value.name)
        elif "scope" in self.attributes:
            del self.attributes["scope"]

    @property
    def signature(self) -> str:
        scope = self.scope
        name = self.name
        type_info = self.type

        if not name or not any(name):
            raise ValueError("Function node does not have a name.")

        out = StringIO()

        if scope == FunctionScope.Class:
            out.write("@classmethod\n")
        elif scope == FunctionScope.Static:
            out.write("@staticmethod\n")

        out.write("def ")
        out.write(name)
        out.write("(")

        arg_text = []

        if scope == FunctionScope.Instance:
            arg_text.append("self")
        elif scope == FunctionScope.Class:
            arg_text.append("cls")

        args = self.arguments

        for arg in args:
            buffer = [arg.name]

            arg_type = arg.type

            buffer.append(": ")
            buffer.append(self.localise_name(arg_type) if arg_type else "typing.Any")

            default = arg.default

            if default:
                buffer.append(" = ")
                buffer.append(default)

            arg_text.append("".join(buffer))

        if any(arg_text):
            out.write(", ".join(arg_text))

        out.write(") -> ")
        out.write(self.localise_name(type_info) if type_info else "None")
        out.write(":")

        return out.getvalue()

    def create_ref(self, simple: bool = False) -> Optional[Reference]:
        name = self.full_name

        if not name:
            return None

        if simple:
            name = "~" + name

        return FunctionRef(text=name)


class DocString(TextElement):
    tagname = "docstring"


class Import(TextElement):
    tagname = "import"

    @property
    def module(self) -> Optional[str]:
        return self.attributes["module"] if self.hasattr("module") else None

    @module.setter
    def module(self, value: Optional[str]) -> None:
        if value:
            self.attributes["module"] = value
        elif "module" in self.attributes:
            del self.attributes["module"]

    @property
    def types(self) -> Sequence[str]:
        if not self.hasattr("types"):
            return ()

        value = str(self.attributes["types"]).split(",")

        return tuple(map(lambda v: v.strip(), value))

    @types.setter
    def types(self, value: Sequence[str]) -> None:
        if value and any(value):
            self.attributes["types"] = ", ".join(value)
        elif "types" in self.attributes:
            del self.attributes["types"]

    def astext(self) -> str:
        module = self.module
        types = self.types

        if any(types):
            return f"from {module} import {', '.join(types)}"
        else:
            return " ".join(("import", module))


class Argument(Typed, Named, Element):
    tagname = "argument"

    @property
    def default(self) -> Optional[str]:
        return self.attributes["default"] if self.hasattr("default") else None

    @default.setter
    def default(self, value: Optional[str]) -> None:
        if value:
            self.attributes["default"] = value
        elif "default" in self.attributes:
            del self.attributes["default"]


class Reference(Inline, TextElement):
    tagname = "ref"

    @property
    def target(self) -> str:
        name = super().astext()

        if name.startswith("~"):
            target = name[1:]

            return target
        elif name.startswith("!"):
            return name[1:]

        return name

    def astext(self) -> str:
        name = super().astext()
        ref_type = self.tagname[:-3]

        if name.startswith("~"):
            title = name.split(".")[-1]
            target = name[1:]

            return "".join([":", ref_type, ":`", title, " <", target, ">`"])
        elif name.startswith("!"):
            return name[1:]

        return "".join([":", ref_type, ":`", name, "`"])


class PythonRef(Reference, ABC):
    pass


class ClassRef(PythonRef):
    tagname = "classref"


class ModuleRef(PythonRef):
    tagname = "modref"


class FunctionRef(PythonRef):
    tagname = "funcref"


class DataRef(PythonRef):
    tagname = "dataref"


class PropertyRef(PythonRef):
    tagname = "propref"


class MethodRef(PythonRef):
    tagname = "methref"


class AttributeRef(PythonRef):
    tagname = "attrref"



import ast
import re
from abc import ABC, abstractmethod
from ast import AST, FunctionDef, arguments
from dataclasses import dataclass
from itertools import chain
from typing import Any, List, Mapping, Optional, OrderedDict, Sequence, cast

from docutils import nodes
from docutils.nodes import Element, Node, field_list, paragraph
from docutils.parsers.rst import Directive
from docutils.transforms import Transform

from bpystubgen.nodes import APIMember, Argument, Class, ClassRef, Data, DocString, Function, FunctionScope, \
    Module, Property
from bpystubgen.parser import parse_type


class ModuleTransform(Transform):
    default_priority = 1

    def apply(self, **kwargs: Any) -> None:
        siblings = tuple(self.startnode.traverse(include_self=False, siblings=True, ascend=True))

        if not any(siblings):
            return

        module = siblings[0]

        def is_class_member(m: APIMember) -> bool:
            p = m.parent
            while p:
                if isinstance(p, Class):
                    return True
                p = p.parent

            return False

        if isinstance(module, Module):
            members = filter(lambda s: isinstance(s, APIMember) and not is_class_member(s), siblings)

            module.parent.remove(module)

            for member in tuple(members):
                ref = member.create_ref()

                if ref:
                    para = paragraph()
                    para += member.create_ref()

                    member.replace_self(para)
                else:
                    member.parent.remove(member)

                module += member

            self.startnode.parent.remove(self.startnode)

            if any(self.document.children):
                docstring = DocString()

                for child in tuple(self.document.children):
                    child.parent.remove(child)
                    docstring += child

                module.insert(0, docstring)

            self.document += module
        else:
            reporter = self.document.reporter
            msg = reporter.error("Missing module element.", base_node=self.document)

            assert isinstance(self.startnode, Element)

            self.startnode.replace_self(msg)


class ModuleDirective(Directive):
    has_content = True

    required_arguments = 1

    def run(self) -> List[Node]:
        name = self.arguments[0]

        elem = Module(name=name)

        # noinspection PyTypeChecker
        pending = nodes.pending(ModuleTransform)
        pending.details.update(self.options)

        document = self.state_machine.document
        document.note_pending(pending)

        return [pending, elem]


@dataclass
class DocStringInfo:
    docstring: DocString

    fields: Mapping[str, str]

    fields_list: Optional[field_list]

    remainder: Sequence[APIMember]


class APIMemberDirective(Directive, ABC):

    def parse_docstring(self) -> DocStringInfo:
        docstring = DocString()
        members = []

        self.state.nested_parse(self.content, self.content_offset, docstring)

        for member in tuple(filter(lambda c: isinstance(c, APIMember), docstring.children)):
            members.append(cast(APIMember, member))
            docstring.remove(member)

        fields = dict()
        fields_elem: Optional[field_list] = None

        if any(docstring.children):
            try:
                fields_elem = next(iter(docstring.traverse(field_list, ascend=False)))

                for field in fields_elem.children:
                    (f_name, f_body) = cast(Element, field).children

                    fields[f_name.astext().strip()] = f_body.astext().strip()

                docstring.remove(fields_elem)
            except StopIteration:
                pass

        return DocStringInfo(docstring, fields, fields_elem, members)


class DataLikeDirective(APIMemberDirective, ABC):
    has_content = True

    required_arguments = 1

    @abstractmethod
    def create_elem(self, name: str) -> APIMember:
        pass

    def run(self) -> List[Node]:
        ds = self.parse_docstring()

        elem = self.create_elem(self.arguments[0].strip())
        elems = [elem]

        if any(ds.docstring.children):
            elem.insert(0, ds.docstring)

        if "type" in ds.fields:
            type_info = parse_type(ds.fields["type"])

            if type_info:
                elem.type = type_info
            else:
                elem.type = "typing.Any"

                document = self.state_machine.document
                reporter = document.reporter

                msg = reporter.warning(
                    f"Data {elem.name} has invalid type: {ds.fields['type']}", base_node=self.state.parent)

                elems.append(msg)
        else:
            elem.type = "typing.Any"

        return elems


class DataDirective(DataLikeDirective):

    def create_elem(self, name: str) -> APIMember:
        return Data(name=name)


class PropertyDirective(DataLikeDirective):

    def create_elem(self, name: str) -> APIMember:
        return Property(name=name)


class FunctionLikeDirective(APIMemberDirective, ABC):
    has_content = True

    required_arguments = 1

    final_argument_whitespace = True

    _optional_func_pattern = re.compile("^(\\w+)\\s*\\(\\s*\\[?([\\w\\s,]*)(\\[[^)]+])]?\\)$")

    @classmethod
    def parse_func(cls, line: str) -> FunctionDef:
        line = str(line).strip()

        match = cls._optional_func_pattern.match(line)

        if match:
            required = map(str, filter(any, map(lambda s: s.strip(), match.group(2).split(","))))
            optional = map(
                lambda a: a + "=None",
                filter(any, map(lambda a: a.strip(), match.group(3).replace("[", "").replace("]", "").split(","))))

            args = chain(required, optional)
            line = "".join([match.group(1), "(", ", ".join(args), ")"])

        source = "".join(["def ", line, "\n" if line.endswith(":") else ":\n", "   ...\n"])

        tree = ast.parse(source)

        return cast(FunctionDef, tree.body[0])

    def parse_args(self, args: arguments, fields: Mapping[str, str]) -> (OrderedDict[str, Argument], Sequence[Node]):
        elems = OrderedDict[str, Argument]()
        messages = []

        count = len(args.args)
        offset = count - len(args.defaults)

        for i in range(count):
            arg = args.args[i]
            default: Optional[AST] = args.defaults[i - offset] if i >= offset else None

            if arg.arg == "self":
                continue

            elem = Argument(name=arg.arg)

            key = f"type {elem.name}"

            if key in fields:
                type_info = parse_type(fields[key])

                if type_info:
                    elem.type = type_info
                else:
                    elem.type = "typing.Any"

                    document = self.state_machine.document
                    reporter = document.reporter

                    msg = reporter.warning(f"Invalid argument type: {fields[key]}", base_node=self.state.parent)
                    messages.append(msg)

            if default:
                elem.default = ast.unparse(default)

            elems[elem.name] = elem

        return elems, messages


class FunctionDirective(FunctionLikeDirective):

    def run(self) -> List[Node]:
        document = self.state_machine.document
        reporter = document.reporter

        ds = self.parse_docstring()

        elems = []

        for line in filter(any, self.arguments[0].replace("\\\n", " ").split("\n")):
            try:
                func = self.parse_func(str(line))

                elem = Function(name=func.name)

                if self.name == "classmethod":
                    elem.scope = FunctionScope.Class
                elif self.name == "staticmethod":
                    elem.scope = FunctionScope.Static
                elif self.name == "method":
                    elem.scope = FunctionScope.Instance
                else:
                    elem.scope = FunctionScope.Module

                if "rtype" in ds.fields:
                    type_info = parse_type(ds.fields["rtype"])

                    if type_info:
                        elem.type = type_info
                    else:
                        elem.type = "typing.Any"

                        msg = reporter.warning(
                            f"Function {func.name} has invalid return type: {ds.fields['rtype']}",
                            base_node=self.state.parent)

                        elems.append(msg)

                if "return" in ds.fields:
                    elem.returns = ds.fields["return"]

                if any(ds.docstring.children):
                    elem.insert(0, ds.docstring)

                (args, messages) = self.parse_args(func.args, ds.fields)

                for arg in args.values():
                    elem += arg

                elems.append(elem)
                elems.extend(messages)
            except SyntaxError:
                msg = reporter.error(f"Invalid function signature: {line}", base_node=self.state.parent)
                elems.append(msg)

        return elems


class ClassDirective(FunctionLikeDirective):

    def run(self) -> List[Node]:
        signature = self.arguments[0].strip()

        ds = self.parse_docstring()

        elem: Optional[Class] = None
        elems = []

        if "(" in signature:
            parent = cast(Element, self.state.parent)

            try:
                func = self.parse_func(signature)

                name = func.name
                elem = Class(name=name)

                if any(parent.children):
                    last_sibling = parent.children[-1]

                    if last_sibling and last_sibling.astext().startswith("base class"):
                        base_types = set(map(lambda t: t.target, last_sibling.traverse(ClassRef, descend=True)))
                        elem.base_types = base_types

                (args, messages) = self.parse_args(func.args, ds.fields)

                elems.extend(messages)

                # Ignore when base classes are used instead of constructor arguments.
                if any(ds.fields):
                    ctor = Function(name="__init__", type="None")

                    ctor.scope = FunctionScope.Instance

                    if ds.fields_list:
                        docstring = DocString()
                        docstring += ds.fields_list

                        ctor += docstring

                    for arg in args.values():
                        ctor += arg

                    elem.insert(0, ctor)
                elif any(args) and not any(elem.base_types):
                    elem.base_types = set(args)

                if any(ds.docstring.children):
                    elem.insert(0, ds.docstring)
            except SyntaxError:
                document = self.state_machine.document
                reporter = document.reporter

                msg = reporter.error(f"Invalid class signature: {signature}", base_node=parent)

                elems.append(msg)
        else:
            name = signature.strip()
            elem = Class(name=name)

            if any(ds.docstring.children):
                elem.insert(0, ds.docstring)

        if elem:
            elems.append(elem)

            for member in ds.remainder:
                elem += member

        return elems


class CurrentModuleDirective(Directive):
    has_content = True

    required_arguments = 1

    def run(self) -> List[Node]:
        return []



from docutils.parsers.rst.directives import register_directive
from docutils.parsers.rst.roles import GenericRole, register_local_role

from bpystubgen.directives import ClassDirective, CurrentModuleDirective, DataDirective, FunctionDirective, \
    ModuleDirective, PropertyDirective
from bpystubgen.generator import generate
from bpystubgen.nodes import AttributeRef, ClassRef, DataRef, PropertyRef, DocString, Function, FunctionRef, \
    MethodRef, Module, ModuleRef, PythonRef, Reference

register_directive("module", ModuleDirective)
register_directive("data", DataDirective)
register_directive("attribute", DataDirective)
register_directive("property", PropertyDirective)
register_directive("function", FunctionDirective)
register_directive("method", FunctionDirective)
register_directive("classmethod", FunctionDirective)
register_directive("staticmethod", FunctionDirective)
register_directive("class", ClassDirective)
register_directive("currentmodule", CurrentModuleDirective)

_known_roles = (
    ("ref", Reference),
    ("mod", ModuleRef),
    ("class", ClassRef),
    ("func", FunctionRef),
    ("meth", MethodRef),
    ("data", DataRef),
    ("attr", AttributeRef),
    ("property", PropertyRef)
)

for (name, cls) in _known_roles:
    register_local_role(name, GenericRole(name, cls))

    if issubclass(cls, PythonRef):
        register_local_role("py:" + name, GenericRole("py:" + name, cls))



from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from itertools import chain
from logging import Logger
from pathlib import Path
from typing import Dict, Iterable, Optional, Sequence

from docutils.core import publish_doctree
from docutils.frontend import OptionParser, Values
from docutils.io import FileInput, FileOutput
from docutils.nodes import document
from docutils.parsers.rst import Parser
from docutils.utils import new_document
from docutils.writers import Writer
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinxcontrib.builders.rst import RstBuilder

from bpystubgen.nodes import Class, Import, Module
from bpystubgen.writer import StubWriter


@dataclass
class TaskContext:
    dest_dir: Path
    writer: Writer
    settings: Values
    logger: Logger
    total: int
    successful: int = 0
    failed: int = 0

    @property
    def env(self) -> BuildEnvironment:
        # noinspection PyUnresolvedReferences
        return self.settings.env

    @property
    def done(self) -> int:
        return self.successful + self.failed

    @property
    def remaining(self) -> int:
        return self.total + self.done


class Task:

    def __init__(self, name: str = "", parent: Optional[Task] = None) -> None:
        self.name = name
        self.parent = parent
        self.children: Dict[str, Task] = dict()
        self.source: Optional[Path] = None
        self.doctree: Optional[document] = None

    @property
    def full_name(self) -> str:
        segments = list(filter(any, map(lambda a: a.name, self.ancestors)))
        segments.append(self.name)

        return ".".join(segments)

    @property
    def total(self) -> int:
        return (1 if self.parent else 0) + sum(map(lambda c: c.total, self.children.values()))

    @property
    def ancestors(self) -> Iterable[Task]:
        if self.parent:
            return chain(self.parent.ancestors, (self.parent,))
        else:
            return iter(())

    @property
    def is_module(self) -> bool:
        return not self.is_class

    @property
    def is_class(self) -> bool:
        return self.source and str.isupper(self.name[0])

    @property
    def has_submodule(self) -> bool:
        return any(filter(lambda c: c.is_module, self.children))

    def resolve(self, path: Sequence[str]) -> Task:
        count = len(path)

        if count == 0:
            raise ValueError("Path cannot be empty")

        name = path[0]

        if name in self.children:
            child = self.children[name]
        else:
            child = Task(name, parent=self)

            self.children[name] = child

        if count == 1:
            return child

        return child.resolve(path[1:])

    def run(self, context: TaskContext) -> None:
        for child in self.children.values():
            child.run(context)

        if not self.parent:
            context.successful += 1
            return

        context.logger.info("Processing %s (%d of %d)", self.full_name, context.done + 1, context.total)

        try:
            if self.source:
                source_path = str(self.source)

                context.logger.debug("Parsing document: %s", source_path)

                fin = FileInput(source_path=source_path)

                context.env.project.docnames.add(source_path)
                context.env.prepare_settings(source_path)

                self.doctree = publish_doctree(
                    fin,
                    source_class=FileInput,
                    source_path=source_path,
                    settings=context.settings)
            elif self.is_module:
                components = (Parser,)
                settings = OptionParser(components=components).get_default_values()

                self.doctree = new_document("", settings=settings)
                self.doctree += Module(name=self.name)

            if not self.is_module or not (self.source or any(self.doctree.children)):
                context.successful += 1
                return

            if any(self.children):
                parent_dir = Path(context.dest_dir, "/".join(self.full_name.split("."))).resolve()
                target = parent_dir / "__init__.pyi"
            else:
                parent_dir = Path(context.dest_dir, "/".join(self.full_name.split(".")[:-1])).resolve()
                target = parent_dir / (self.name + ".pyi")

            context.logger.debug("Generating module: %s", target)

            try:
                module = next(iter(self.doctree.traverse(Module)))

                for child in filter(lambda c: c.is_class, self.children.values()):
                    for node in child.doctree.traverse(Class):
                        node.parent.remove(node)
                        module += node

                module.import_types()
                module.sort_members()

                index = 1 if module.docstring else 0

                for child in filter(lambda c: c.is_module, self.children.values()):
                    for node in child.doctree.traverse(Module):
                        module.insert(index, Import(module=".", types=module.localise_name(node.name)))

                target.parent.mkdir(parents=True, exist_ok=True)

                output_path = str(target)

                context.logger.debug("Generating stub file: %s", output_path)

                marker = target.parent / "py.typed"

                with open(marker, "w"):
                    pass

                context.writer.write(self.doctree, FileOutput(destination_path=output_path))
                context.writer.assemble_parts()
            except StopIteration:
                context.logger.warning("Parsed doctree does not contain any module: %s", self.full_name)

            context.successful += 1
        except BaseException as e:
            context.failed += 1

            context.logger.error("Failed to process task %s", self.name, exc_info=e)


def generate(src_dir: Path, dest_dir: Path, log_level: int = logging.INFO) -> None:
    logging.basicConfig(level=log_level, format="[%(levelname)s] %(name)s - %(message)s")

    logger = logging.getLogger("bpystubgen")

    logger.info("Reading *.rst files from the source location: %s", src_dir)

    started = time.perf_counter()

    root = Task()

    for file in sorted(src_dir.rglob("*.rst")):
        segments = file.name.split(".")[:-1]

        task = root.resolve(segments)
        task.source = file

    # noinspection DuplicatedCode
    components = (Parser,)

    app = Sphinx(srcdir=".", confdir=None, outdir=str(dest_dir), doctreedir=".", buildername="text")

    settings = OptionParser(components=components).get_default_values()

    settings.line_length_limit = 15000
    settings.report_level = 5
    settings.traceback = True
    settings.env = app.env

    builder = RstBuilder(app)
    builder.config.rst_indent = 2

    writer = StubWriter(builder)

    context = TaskContext(dest_dir, writer, settings, logger, root.total)

    root.run(context)

    logger.info(
        "Finished processing %d entries in %d seconds (successful: %d, failed: %d).",
        context.total,
        time.perf_counter() - started,
        context.successful,
        context.failed)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="bpystubgen",
    version="0.1.7",
    author="Xavier Cho",
    author_email="mysticfallband@gmail.com",
    description="A utility to generate Python API stubs from documentation files in reStructuredText format.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mysticfall/bpystubgen",
    packages=["bpystubgen"],
    install_requires=["docutils==0.17.1", "sphinxcontrib-restbuilder==0.3", "black==21.8b0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ]
)
